/*
 * Copyright (c) 1987, Fujitsu LTD. (Itaru ICHIKAWA).
 * Copyright (c) 1996-2009, The nkf Project.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 * claim that you wrote the original software. If you use this software
 * in a product, an acknowledgment in the product documentation would be
 * appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 * misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 */
#define WINMAIN
#define printf msgboxprintf
#define fprintf dllprintf

/*
  define USEHELP   not expand wildcard "/?"
  undef  USEHELP   expand wildcard "/?" as "\a","\b",...
  define WIN32     for win32ap
  undef  WIN32     for dos
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <fcntl.h>
#include <io.h>
#ifdef LSI_C
#include <jctype.h>
#else
#ifdef WIN32
#include <windows.h>
#define iskanji(x) IsDBCSLeadByte(x)
#else
#define iskanji(x) ( ((x) >= 0x81 && (x) <= 0x9f) || ((x) >= 0xe0 && (x) <= 0xfc) )
/* SBCS  #define iskanji 0 */
#endif
#endif

#ifdef WINMAIN
char msgbuff[4096];
UINT uType = MB_OK;

int winputs(char *p)
{
    int sts;

    sts = MessageBox(NULL,p,GetCommandLine(),uType);
    uType = MB_OK; 
    return sts;
}

int msgboxprintf(const char *fmt, ...)
{
    va_list argp;
    int ret;
    int sts;

    if ( fmt != NULL && *fmt != 0 ) {
        va_start(argp, fmt);
        ret = _vsnprintf(msgbuff,sizeof msgbuff - 1,fmt,argp);
        va_end(argp);
    }
    sts = MessageBox(NULL,msgbuff,GetCommandLine(),MB_OKCANCEL);
    if ( sts == IDCANCEL ) {
        exit(1);
    }
    return ret;
}

char *ubuff = msgbuff;
int ulen = sizeof msgbuff;
int uret = TRUE;

int dllprintf(FILE *fp,const char *fmt,...)
{
    va_list argp;
    int sts;

/*fputs(fmt,ofp);*/
    if ( uret != FALSE && ulen >= 1 && fmt != NULL && *fmt != 0 ) {
        va_start(argp, fmt);
        sts = _vsnprintf(ubuff,ulen - 1,fmt,argp);
        va_end(argp);
        if ( sts >= 0 ) {
            ubuff += sts;
            ulen -= sts;
        } else {
            uret = FALSE;
        }
/*fputs(ubuff,ofp);*/
        return sts;
    } else return 0;
}

int dllstdprintf(char *fmt,...)
{
    va_list argp;
    int sts;

    if ( uret != FALSE && ulen >= 1 && fmt != NULL && *fmt != 0 ) {
        va_start(argp, fmt);
        sts = _vsnprintf(ubuff,ulen - 1,fmt,argp);
        va_end(argp);
        if ( sts >= 0 ) {
            ubuff += sts;
            ulen -= sts;
        } else {
            uret = FALSE;
        }
        return sts;
    } else return 0;
}
#endif

void dbgprintf(char *fmt,...)
{
    va_list argp;
    int sts;

    if ( uret != FALSE && ulen >= 1 && fmt != NULL && *fmt != 0 ) {
        va_start(argp, fmt);
        sts = _vsnprintf(msgbuff,sizeof msgbuff,fmt,argp);
        va_end(argp);
    }
    OutputDebugString(msgbuff);
}

extern int globmain(int,char **);

static char *pgname;

static char *dirname(char *p)
{
    char *s;
    char *q;
    int c;

    if ( (p = s = strdup(p)) == NULL ) {
        msgboxprintf("%s: inadequate memory.\n",pgname);
        exit(1);
    }
    if ( (*p == '\\' || *p == '/') && (*(p + 1) == '\\' || *(p + 1) == '/') ) { /* \\xxx */
        p += 2;
    } else if ( (*p == '\\' || *p == '/') ) { /* \xxx */
        p += 1;
    } else if ( isalpha(*p) && *(p + 1) == ':' && (*(p + 2) == '\\' || *(p + 2) == '/') ) { /* a:\xxx */
        p += 3;
    } else if ( isalpha(*p) && *(p + 1) == ':' ) { /* a:xxx */
        p += 2;
    }
    q = p;
    while ( (c = *p) != '\0' ) {
       if ( (c == '\\'  || c == '/') ) {
           q = p;
       } else if ( iskanji((BYTE)c) ) {
           if ( *(p + 1) != '\0' ) {
               p++;
           /*
           } else {
               dbgprintf("%s: %s: bad filename.\n",pgname,s);
               exit (1);
           */
           }
       }
       p++;
    } 
    *q = '\0';
    return s;
}

static char *newname(char *dir,char *name)
{
    int n;
    char *p;


    n = strlen(name);
    if ( (strcmp(dir,"\\") == 0 || strcmp(dir,"/") == 0) ) { /* "\" -> "\name" */
        p = malloc(n + 2);
        sprintf(p,"\\%s",name);
    } else if ( isalpha(*dir) && *(dir + 1) == ':' &&
                *(dir + 2) == '\0') { /* "a:"  -> "a:name" */
        p = malloc(n + 3);
        sprintf(p,"%s%s",dir,name);
    } else if ( isalpha(*dir) && *(dir + 1) == ':' &&
                (*(dir + 2) == '\\' || *(dir + 2) == '/') && *(dir + 3) == '\0') { /* "a:\"  -> "a:\name" */
        p = malloc(n + 4);
        sprintf(p,"%s%s",dir,name);
    } else if ( *dir == 0 ) {
        p = malloc(n + 1);
        sprintf(p,"%s",name);
    } else {
        p = malloc(strlen(dir) + n + 2);
        sprintf(p,"%s\\%s",dir,name);
    }
    return p;
}

static int ckglob(char *p)
{
    int c;

#ifdef USEHELP
    if ( ((c = *p) == '/' || c == '-') && *(p+1) == '?' && *(p+2) == 0 ) return 0;
#endif
    while( (c = *p) != 0 ) {
        if ( c == '*' || c == '?' ) {
            return 1;
        } else if ( iskanji((BYTE)c) ) {
            if ( *(p + 1) != 0 ) p++;
        }
        p++;
    }
    return 0;
}

#define main globmain
#include "nkf.c"
#undef main

int xmain(int argc,char *argv[])
{
    int sts;
    int i;
    int g;
    int bad;
    char *dir;
#ifdef WIN32
    WIN32_FIND_DATA FindFileData;
    HANDLE hFind;
#else
    struct find_t FindFileData;
#endif
    int newargc;
    char **newargv;
    struct arglink {
      struct arglink *next;
      char *arg;
    } *linkp,*p,*q;

/*
    dbgprintf("main:%d\n",argc);
    for ( i = 0 ; i < argc ; i++ ) {
        dbgprintf("%d %s\n",i,argv[i]);
    }
*/
    pgname = argv[0];
    g = 0;
    for ( i = 1 ; i < argc ; i++ ) {
        if ( ckglob(argv[i]) ) {
            g = 1;
            break;
        }
    }
    if ( g == 0 ) {
        sts = globmain(argc,argv);
    } else {
        linkp = malloc(sizeof(struct arglink));
        q = linkp;
        q->arg = argv[0];
        newargc = 1;
        for ( i = 1 ; i < argc  ; i++ ) {
            if ( !ckglob(argv[i]) ) {
/*dbgprintf("%d: [%s] !ckglob\n",i,argv[i]);*/
                p = malloc(sizeof(struct arglink));
                p->arg = argv[i];
                q->next = p;
                q = p;
                newargc++;
            } else {
/*dbgprintf("%d: [%s] ckglob ",i,argv[i]);*/
#ifdef WIN32
                if ( (hFind = FindFirstFile(argv[i],&FindFileData)) == INVALID_HANDLE_VALUE )
#else
                if ( _dos_findfirst(argv[i],_A_NORMAL | _A_HIDDEN | _A_SYSTEM | _A_SUBDIR,&FindFileData) != 0 )
#endif
                {
/*dbgprintf(" NOT match \n");*/
                    p = malloc(sizeof(struct arglink));
                    p->arg = argv[i];
                    q->next = p;
                    q = p;
                    newargc++;
/*dbgprintf("%d: [%s] argument not matched.\n",i,p->arg);*/
                } else {
                    bad = 1;
/*dbgprintf(" match ");*/
                    dir = dirname(argv[i]);
/*dbgprintf(" dirname=%s FindFileData.cFileName=%s \n",dir,FindFileData.cFileName);*/
                    do {
#ifdef WIN32
                        if ( strcmp(FindFileData.cFileName,".") == 0 || strcmp(FindFileData.cFileName,"..") == 0 ) continue;
#else /*WIN32*/
                        if ( strcmp(FindFileData.name,".") == 0 || strcmp(FindFileData.name,"..") == 0 ) continue;
#endif /*WIN32*/
                        bad = 0;
                        p = malloc(sizeof(struct arglink));
#ifdef WIN32
                        p->arg = malloc(strlen(FindFileData.cFileName) + 1);
                        strcpy(p->arg,newname(dir,FindFileData.cFileName));
#else /*WIN32*/
                        p->arg = malloc(strlen(FindFileData.name) + 1);
                        strcpy(p->arg,newname(dir,FindFileData.name));
#endif /*WIN32*/
                        q->next = p;
                        q = p;
                        newargc++;
                    }
#ifdef WIN32
                      while ( FindNextFile(hFind,&FindFileData) != 0 );
#else /*WIN32*/
                      while ( _dos_findnext(&FindFileData) == 0 );
#endif /*WIN32*/
                    if ( bad ) {
/*dbgprintf("%d: %s: BAD matched.\n",i,argv[i]);*/
                        p = malloc(sizeof(struct arglink));
                        p->arg = argv[i];
                        q->next = p;
                        q = p;
                        newargc++;
                    }
                    free(dir);
                }
#ifdef WIN32
                FindClose(hFind);
#else /*WIN32*/
#ifdef FINDCLOSE
                _dos_findclose(&FindFileData);
#endif
#endif /*WIN32*/
            }
        }
        newargv = malloc((newargc + 1) * sizeof(char *));
        p = linkp;
        for ( i = 0 ; i < newargc ; i++ ) {
            newargv[i] = p->arg;
            p = p->next;
        }
        newargv[newargc] = NULL;
/*
for ( i = 0 ; i < newargc ; i++ ) {
    dbgprintf("%d:: [%s]\n",i,newargv[i]);
}
*/
        sts = globmain(newargc,newargv);
     /* free() linkp without argv[*] */
     /* free(newargv); */
    }
    return sts;
}

#ifndef WIN32
/* dos */
extern char *getcmdline(char *cmdline);
#endif /*WIN32*/

#ifdef WINMAIN
int WINAPI WinMain(HINSTANCE hInstance,// ݂̃CX^X̃nh
  HINSTANCE hPrevInstance,  // ȑÕCX^X̃nh
  LPSTR lpCmdLine,          // R}hC
  int nCmdShow              // \
)
{
    int i;
    int sts;
    char *p;
    char *q;
    int ec;
    int argc;
#ifndef WIN32
    char *argv[64];
#define MAXARGV 128
    char cmdline[MAXARGV];

    p = getcmdline(cmdline);
#else /*WIN32*/
#define MAXARGV 16384
    char *argv[MAXARGV];

    p = strdup(GetCommandLine());
/*
dbgprintf("GetCommandLine()=%s\n",p);
*/
    if ( p == NULL ) {
        msgboxprintf("out of memory\n");
        exit(1);
    }
    if ( *p == 0 ) {
        ;
    } else if ( *p == '"' ) {
        ec = '"';
        p++;
        argv[0] = p;
    } else {
        ec = ' ';
        argv[0] = p;
    }
    while ( *p ) {
        if ( *p == ec ) {
            if ( *p == ' ' ) {
                *p = 0;
                p++;
                break;
            } else {
                *p = 0;
                p++;
                break;
            }
        }
        p++;
    }
#endif /*WIN32*/
    if ( *p == ' ' ) p++;
    argc = 1;
    while ( *p ) {
        if ( argc >= MAXARGV ) {
            msgboxprintf("too many arguments\n");
            exit(1);
        }
        if ( *p == 0 ) {
            ;
        } else if ( *p == '"' ) {
            ec = '"';
            p++;
            argv[argc++] = p;
        } else {
            ec = ' ';
            if ( *p == '\\' && *(p+1) == '"' ) p++;
            argv[argc++] = p;
            p++;
        }
        while ( *p ) {
            if ( ec  == ' ' ) {
                if ( *p == ' ' ) {
                    *p = 0;
                    p++;
                    break;
                } else if ( *p == '"' )  {
                    memmove(p,p+1,strlen(p));
                    q = strchr(p,'"');
                    if ( q == NULL ) {
                        p += strlen(p);
                        break;
                    } else {
                        p = q;
                        memmove(p,p+1,strlen(p));
                    }
                } else if ( *p == '\\' && *(p+1) == '"' )  {
                    memmove(p,p+1,strlen(p));
                }
                p++;
            } else if ( ec  == '"' ) {
                if ( *p == '"' )  {
                    *p = 0;
                    p++;
                    break;
                } else if ( *p == '\\' && *(p+1) == '"' )  {
                    memmove(p,p+1,strlen(p));
                }
                p++;
            }
        }
        while ( *p == ' ' ) p++;
    }
    argv[argc] = NULL;
    ubuff = msgbuff;
    ulen = sizeof msgbuff;
    uret = TRUE;
/*
dbgprintf("argc=%d\n",argc);
for ( i = 0 ; i < argc ; i++ ) {
    dbgprintf("%d:: [%s]\n",i,argv[i]);
}
*/
    if ( argc >= 2 ) {
        if ( strcmp(argv[1],"/?") == 0|| strcmp(argv[1],"--help") == 0 ) {
            usage();
            MessageBox(NULL,msgbuff,GetCommandLine(),uType);
            exit(1);
        } else if ( strcmp(argv[1],"-v") == 0 || strcmp(argv[1],"--version") == 0 ) {
            version();
            MessageBox(NULL,msgbuff,GetCommandLine(),uType);
            exit(1);
        } else if ( strcmp(argv[1],"-V") == 0 ) {
            show_configuration();
            MessageBox(NULL,msgbuff,GetCommandLine(),uType);
            exit(1);
        }
    }
    sts = xmain(argc,argv);
/*
    dbgprintf("msgbuff [%s]\n",msgbuff);
*/
    return sts;
}
#endif
